/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.propertysheet.reference;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eclipse.ui.views.properties.ComboBoxPropertyDescriptor;
import org.eclipse.ui.views.properties.IPropertyDescriptor;
import org.eclipse.ui.views.properties.IPropertySource;
import org.eclipse.ui.views.properties.PropertyDescriptor;
import org.eclipse.ui.views.properties.TextPropertyDescriptor;

import eu.etaxonomy.cdm.common.CdmUtils;
import eu.etaxonomy.cdm.model.agent.Team;
import eu.etaxonomy.cdm.model.agent.TeamOrPersonBase;
import eu.etaxonomy.cdm.model.common.TimePeriod;
import eu.etaxonomy.cdm.model.reference.Article;
import eu.etaxonomy.cdm.model.reference.BibtexReference;
import eu.etaxonomy.cdm.model.reference.Book;
import eu.etaxonomy.cdm.model.reference.BookSection;
import eu.etaxonomy.cdm.model.reference.CdDvd;
import eu.etaxonomy.cdm.model.reference.Database;
import eu.etaxonomy.cdm.model.reference.Generic;
import eu.etaxonomy.cdm.model.reference.InProceedings;
import eu.etaxonomy.cdm.model.reference.Journal;
import eu.etaxonomy.cdm.model.reference.Map;
import eu.etaxonomy.cdm.model.reference.Patent;
import eu.etaxonomy.cdm.model.reference.PersonalCommunication;
import eu.etaxonomy.cdm.model.reference.PrintSeries;
import eu.etaxonomy.cdm.model.reference.PrintedUnitBase;
import eu.etaxonomy.cdm.model.reference.Proceedings;
import eu.etaxonomy.cdm.model.reference.PublicationBase;
import eu.etaxonomy.cdm.model.reference.ReferenceBase;
import eu.etaxonomy.cdm.model.reference.Report;
import eu.etaxonomy.cdm.model.reference.SectionBase;
import eu.etaxonomy.cdm.model.reference.StrictReferenceBase;
import eu.etaxonomy.cdm.model.reference.Thesis;
import eu.etaxonomy.cdm.model.reference.WebPage;
import eu.etaxonomy.taxeditor.ITaxEditorConstants;
import eu.etaxonomy.taxeditor.model.CdmUtil;
import eu.etaxonomy.taxeditor.propertysheet.TimePeriodPropertySource;
import eu.etaxonomy.taxeditor.propertysheet.YearValidator;

/**
 * Supplies the properties of a <code>ReferenceBase</code> object for display by
 * a <code>PropertySheet</code>.
 * <p>
 * Includes a drop-down menu that allows the user to change <code>ReferenceBase</code> 
 * subclasses. This causes all fields except <code>TitleCache</code> to be wiped clean.
 * <p>
 * Normally, this would be called from another <code>IPropertySource</code>  
 * as follows:
 * <pre>
 * public Object getPropertyValue(Object id) {
 * 
 * 		...
 * 
 * 		if (id.equals(P_ID_BIBREF)) {
 *			ReferenceBase citation = descriptionElement.getCitation();
 *			ReferencePropertySource bibRefPropertySource = new ReferencePropertySource(citation);
 *			bibRefPropertySource.addPropertyChangeListener(new PropertyChangeListener() {
 *				@Override
 *				public void propertyChange(PropertyChangeEvent evt) {
 *					descriptionElement.setCitation((ReferenceBase) evt.getNewValue());
 *				}
 *			});
 *			return bibRefPropertySource;
 *		}		
 * </pre>
 * <p>
 * Note: if <code>P_ID_BIBREF</code> in the above example is associated with an editable 
 * <code>TextPropertyDescriptor</code>, it will initalize its cell editor by calling 
 * <code>getEditableValue()</code>. If associated with a non-editable <code>PropertyDescriptor</code>, 
 * it will get its value from <code>getString()</code>.
 * @author p.ciardelli
 * @created 03.11.2008
 * @version 1.0
 */
public class ReferencePropertySource implements IPropertySource {
	private static final Logger logger = Logger
			.getLogger(ReferencePropertySource.class);

	ReferenceBase reference;
	
    // Property unique keys
	public static final String P_ID_SEARCH = "P_ID_SEARCH";
	public static final String P_ID_EDITABLECACHE = "P_ID_EDITABLECACHE";	
	public static final String P_ID_REFERENCETYPE = "P_ID_REFERENCETYPE";
	public static final String P_ID_AUTHORTEAM = "P_ID_AUTHORTEAM";
	public static final String P_ID_YEAR = "P_ID_YEAR";
	public static final String P_ID_URI = "P_ID_URI";
	public static final String P_ID_DATEPUBLISHED = "P_ID_DATEPUBLISHED";
	public static final String P_ID_TITLE = "P_ID_TITLE";
	public static final String P_ID_INJOURNAL = "P_ID_INJOURNAL";
	public static final String P_ID_PAGES = "P_ID_PAGES";
	public static final String P_ID_SERIES = "P_ID_SERIES";
	public static final String P_ID_VOLUME = "P_ID_VOLUME";
	public static final String P_ID_EDITOR = "P_ID_EDITOR";
	public static final String P_ID_PLACEPUBLISHED = "P_ID_PLACEPUBLISHED";
	public static final String P_ID_PUBLISHER = "P_ID_PUBLISHER";
	public static final String P_ID_ISSN = "P_ID_ISSN";
	public static final String P_ID_INSTITUTION = "P_ID_INSTITUTION";
	public static final String P_ID_SCHOOL = "P_ID_SCHOOL";
	public static final String P_ID_INSERIES = "P_ID_INSERIES";
	public static final String P_ID_SERIESPART = "P_ID_SERIESPART";
	public static final String P_ID_ISBN = "P_ID_ISBN";
	public static final String P_ID_ORGANIZATION = "P_ID_ORGANIZATION";
	public static final String P_ID_PRINTEDUNIT = "P_ID_PRINTEDUNIT";
	public static final String P_ID_INBOOK = "P_ID_INBOOK";
	public static final String P_ID_INPROCEEDINGS = "P_ID_INPROCEEDINGS";
	public static final String P_ID_BIBTEX_ENTRYTYPE = "P_ID_BIBTEX_ENTRYTYPE";
	public static final String P_ID_JOURNAL = "P_ID_JOURNAL";
	public static final String P_ID_BOOKTITLE = "P_ID_BOOKTITLE";
	public static final String P_ID_CHAPTER = "P_ID_CHAPTER";
	public static final String P_ID_EDITION = "P_ID_EDITION";
	public static final String P_ID_NUMBER = "P_ID_NUMBER";
	public static final String P_ID_ANNOTE = "P_ID_ANNOTE";
	public static final String P_ID_ADDRESS = "P_ID_ADDRESS";
	public static final String P_ID_HOWPUBLISHED = "P_ID_HOWPUBLISHED";
	public static final String P_ID_REPORTTYPE = "P_ID_REPORTTYPE";
	public static final String P_ID_MONTH = "P_ID_MONTH";
	public static final String P_ID_EPRINT = "P_ID_EPRINT";
	public static final String P_ID_NOTE = "P_ID_NOTE";
	public static final String P_ID_CROSSREF = "P_ID_CROSSREF";
	public static final String P_ID_PROTECT_CACHE = "P_ID_PROTECT_CACHE";
		
    // Property display keys
	public static final String P_SEARCH = "001:Search";
	public static final String P_EDITABLECACHE = "002:Editable Cache";
	public static final String P_PROTECT_CACHE = "0020:Protect Cache from overwriting?";
	public static final String P_REFERENCETYPE = "003:Reference Type";
	public static final String P_AUTHORTEAM = "01:Author Team (Cache)";
	public static final String P_YEAR = "02:Year";
	public static final String P_DATEPUBLISHED = "03:Date Published";
	public static final String P_TITLE = "06:Title";
	public static final String P_INJOURNAL = "07:In Journal";
	public static final String P_VOLUME = "10:Volume";
	public static final String P_EDITION = "110:Edition";
	public static final String P_SERIES = "11:Series";
	public static final String P_PAGES = "110:Pages";
	public static final String P_EDITOR = "12:Editor";
	public static final String P_PUBLISHER = "13:Publisher";
	public static final String P_PLACEPUBLISHED = "14:Place Published";
	public static final String P_ISSN = "15:ISSN";
	public static final String P_INSTITUTION = "16:Institution";
	public static final String P_SCHOOL = "17:School";
	public static final String P_INSERIES = "18:In Series";
	public static final String P_SERIESPART = "19:Series Part";
	public static final String P_ISBN = "20:ISBN";
	public static final String P_URI = "201:URI";
	public static final String P_ORGANIZATION = "21:Organization";
	public static final String P_PRINTEDUNIT = "22:Printed Unit";
	public static final String P_INBOOK = "23:In Book";
	public static final String P_INPROCEEDINGS = "24:In Proceedings";
	public static final String P_BIBTEX_ENTRYTYPE = "25:BibTeX Entry Type";
	public static final String P_JOURNAL = "26:Journal";
	public static final String P_BOOKTITLE = "27:Book Title";
	public static final String P_CHAPTER = "28:Chapter";
	public static final String P_NUMBER = "30:Number";
	public static final String P_ANNOTE = "31:Annote";
	public static final String P_ADDRESS = "32:Address";
	public static final String P_HOWPUBLISHED = "33:How Published";
	public static final String P_REPORTTYPE = "34:Report Type";
	public static final String P_MONTH = "35:Month";
	public static final String P_EPRINT = "36:E-Print";
	public static final String P_NOTE = "37:Note";
	public static final String P_CROSSREF = "38:BibTeX Crossref";		
	
	protected static HashMap<Class, String> referenceTypeMap = null;
	
	private static final String[] P_BIBTEX_ENTRYTYPE_MENU = new String[] {"ARTICLE", "BOOK", "BOOKLET", "INBOOK", "INCOLLECTION", "PROCEEDINGS", "INPROCEEDINGS", "CONFERENCE", "MANUAL", "MASTERTHESIS", "PHDTHESIS", "TECHREPORT", "UNPUBLISHED", "MISC"};

	private static final int CACHE_NOT_PROTECTED = 0;
	private static final int CACHE_PROTECTED = 1;	
	
	protected void populateReferenceTypes() {
		
		// LinkedHashMap maintains insertion order
		referenceTypeMap = new LinkedHashMap<Class, String>();
		
		referenceTypeMap.put(ReferenceBase.class, "");
//		referenceTypeMap.put(BibtexReference.class, "BibTeX Reference");
		referenceTypeMap.put(Article.class, "Article");
		referenceTypeMap.put(Generic.class, "Generic");
		referenceTypeMap.put(Patent.class, "Patent");
		referenceTypeMap.put(PersonalCommunication.class, "Personal Communication");
		referenceTypeMap.put(CdDvd.class, "CD / DVD");
		referenceTypeMap.put(Database.class, "Database");
		referenceTypeMap.put(Journal.class, "Journal");
		referenceTypeMap.put(Map.class, "Map");
		referenceTypeMap.put(Book.class, "Book");
		referenceTypeMap.put(Proceedings.class, "Proceedings");
		referenceTypeMap.put(PrintSeries.class, "Print Series");
		referenceTypeMap.put(Report.class, "Report");
		referenceTypeMap.put(Thesis.class, "Thesis");
		referenceTypeMap.put(WebPage.class, "Web Page");
		referenceTypeMap.put(BookSection.class, "Book Section");
		referenceTypeMap.put(InProceedings.class, "In Proceedings");
	}
	
	protected String[] getReferenceTypeArray() {
		if (referenceTypeMap == null) {
			populateReferenceTypes();
		}
		return referenceTypeMap.values().toArray(new String[] {});
	}
	
	protected Set<Class> getReferenceClassSet() {
		return referenceTypeMap.keySet();
	}
		
	public ReferencePropertySource(ReferenceBase reference) {
		
		// Default type of ReferenceBase is Generic
		if (reference == null) {
			reference = Generic.NewInstance();
		}		
		this.reference = reference;

		initDescriptors();
	}

	protected void initDescriptors() {
		
		List<String> displayFields = new ArrayList<String>();
	
		// Drop-down menu to change reference type
		displayFields.add(P_ID_REFERENCETYPE);
		
		// ReferenceBase fields
		displayFields.add(P_ID_SEARCH);
		displayFields.add(P_ID_EDITABLECACHE);
		displayFields.add(P_ID_PROTECT_CACHE);
		displayFields.add(P_ID_AUTHORTEAM);
		displayFields.add(P_ID_URI);
		
		Class referenceClass = reference.getClass();
		
		if (reference instanceof BibtexReference) {
						
			displayFields.add(P_ID_BIBTEX_ENTRYTYPE);
			displayFields.add(P_ID_YEAR);
			displayFields.add(P_ID_JOURNAL);
			displayFields.add(P_ID_BOOKTITLE);
			displayFields.add(P_ID_CHAPTER);			
			displayFields.add(P_ID_TITLE);
			displayFields.add(P_ID_SERIES);
			displayFields.add(P_ID_EDITION);
			displayFields.add(P_ID_VOLUME);
			displayFields.add(P_ID_NUMBER);
			displayFields.add(P_ID_PAGES);
			displayFields.add(P_ID_ANNOTE);
			displayFields.add(P_ID_EDITOR);
			displayFields.add(P_ID_INSTITUTION);
			displayFields.add(P_ID_SCHOOL);
			displayFields.add(P_ID_ORGANIZATION);
			displayFields.add(P_ID_PUBLISHER);
			displayFields.add(P_ID_ADDRESS);
			displayFields.add(P_ID_HOWPUBLISHED);
			displayFields.add(P_ID_REPORTTYPE);
			displayFields.add(P_ID_MONTH);
			displayFields.add(P_ID_EPRINT);
			displayFields.add(P_ID_NOTE);
			displayFields.add(P_ID_CROSSREF);
		}
	
		if (reference instanceof StrictReferenceBase) {
			
			displayFields.add(P_ID_DATEPUBLISHED);
			displayFields.add(P_ID_TITLE);
	
			if (referenceClass == Article.class) {
				displayFields.add(P_ID_INJOURNAL);
				displayFields.add(P_ID_PAGES);
				displayFields.add(P_ID_SERIES);
				displayFields.add(P_ID_VOLUME);
			}
	
			if (referenceClass == Generic.class) {
				displayFields.add(P_ID_PAGES);
				displayFields.add(P_ID_SERIES);
				displayFields.add(P_ID_VOLUME);
				displayFields.add(P_ID_EDITOR);
				displayFields.add(P_ID_PLACEPUBLISHED);
				displayFields.add(P_ID_PUBLISHER);
			}			
			
			if (referenceClass == Patent.class) {
				// No additional fields
			}
			
			if (referenceClass == PersonalCommunication.class) {
				// No additional fields				
			}
			if (reference instanceof PublicationBase) {
	
				displayFields.add(P_ID_PLACEPUBLISHED);
				displayFields.add(P_ID_PUBLISHER);
				
				if (referenceClass == CdDvd.class) {
					// No additional fields							
				}
				
				if (referenceClass == Database.class) {
					// No additional fields					
				}
				
				if (referenceClass == Journal.class) {	
					displayFields.add(P_ID_ISSN);
				}
	
				if (referenceClass == Map.class) {
					// No additional fields					
				}				
				
				if (referenceClass == PrintSeries.class) {
					displayFields.add(P_ID_SERIES);					
				}
				
				if (referenceClass == Report.class) {
					displayFields.add(P_ID_INSTITUTION);					
				}
				
				if (referenceClass == Thesis.class) {
					displayFields.add(P_ID_SCHOOL);						
				}
				
				if (referenceClass == WebPage.class) {
					// No additional fields		
				}
				
				if (reference instanceof PrintedUnitBase) {
					
					displayFields.add(P_ID_EDITOR);
					displayFields.add(P_ID_INSERIES);
					displayFields.add(P_ID_PAGES);
					displayFields.add(P_ID_SERIESPART);
					displayFields.add(P_ID_VOLUME);
					
					if (referenceClass == Book.class) {
						displayFields.add(P_ID_ISBN);
						displayFields.add(P_ID_EDITION);
					}
					
					if (referenceClass == Proceedings.class) {
						displayFields.add(P_ID_ORGANIZATION);							
					}
				}
			}
			
			if (reference instanceof SectionBase) {
				
				displayFields.add(P_ID_PAGES);	
				displayFields.add(P_ID_PRINTEDUNIT);
				
				if (referenceClass == BookSection.class) {
					displayFields.add(P_ID_INBOOK);
				}
				
				if (referenceClass == InProceedings.class) {
					displayFields.add(P_ID_INPROCEEDINGS);					
				}
			}
		}
		
		for (String field : displayFields) {
			addDescriptor(field);
		}
	}

	protected Vector<PropertyDescriptor> descriptors = new Vector<PropertyDescriptor>();
	
	protected void addDescriptor(String id) {
		
		// Reference search
		if (id.equals(P_ID_SEARCH)) {
			descriptors.addElement(
					new ReferenceSearchDescriptor(P_ID_SEARCH, P_SEARCH, getSearchType()) {
						protected void saveReference(ReferenceBase reference) {
							setPropertyValue(P_ID_SEARCH, reference);
						}
			});
		}
		
		// Editable cache
		if (id.equals(P_ID_EDITABLECACHE)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_EDITABLECACHE, P_EDITABLECACHE));
		}
		
		// Protect cache?
		if (id.equals(P_ID_PROTECT_CACHE)) {
			descriptors.addElement(
					new ComboBoxPropertyDescriptor(P_ID_PROTECT_CACHE, P_PROTECT_CACHE, 
							new String[] {"no", "yes"}));
		}
		
		// Reference type
		if (id.equals(P_ID_REFERENCETYPE)) {
			descriptors.addElement(createReferenceTypeDescriptor());
		}
		
		// Author team
		if (id.equals(P_ID_AUTHORTEAM)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_AUTHORTEAM, P_AUTHORTEAM));			
		}
		
		// Year
		if (id.equals(P_ID_YEAR)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_YEAR, P_YEAR));
		}
//		if (id.equals(P_ID_YEAR)) {
////			TextPropertyDescriptor yearDescriptor = new TextPropertyDescriptor(P_ID_YEAR, P_YEAR) {
////			    public CellEditor createPropertyEditor(Composite parent) {
////			        CellEditor editor = new TextCellEditor(parent);
////			        
////			        // Colors cell yellow upon activation
////			        editor.getControl().setBackground(new Color(Display.getDefault(), WarningAnnotation.WARNING_RGB));
////			        
////			        if (getValidator() != null) {
////						editor.setValidator(getValidator());
////					}
////			        return editor;
////			    }
////			};
//			
//			TextPropertyDescriptor yearDescriptor = new TextPropertyDescriptor(P_ID_YEAR, P_YEAR);
//			
//			// BibtexReference.setYear() takes a String, no need for validation
//			if (!(reference instanceof BibtexReference)) {
//				yearDescriptor.setValidator(new YearValidator());
//			}
//			descriptors.addElement(yearDescriptor);		
//		}
		
		// Date published
		if (id.equals(P_ID_DATEPUBLISHED)) {
			
			// TimePeriodPropertySource returns TimePeriod.getYear() - make this editable
			TextPropertyDescriptor yearDescriptor = new TextPropertyDescriptor(
					P_ID_DATEPUBLISHED, P_DATEPUBLISHED);
			yearDescriptor.setValidator(new YearValidator());
			descriptors.addElement(yearDescriptor);			
		} 
				
		// URI
		if (id.equals(P_ID_URI)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_URI, P_URI));			
		} 
		
		// Title
		if (id.equals(P_ID_TITLE)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_TITLE, P_TITLE));			
		} 
		
		// In journal
		if (id.equals(P_ID_INJOURNAL)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_INJOURNAL, P_INJOURNAL));			
		}
		
		// Pages
		if (id.equals(P_ID_PAGES)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_PAGES, P_PAGES));			
		} 
		
		// Series
		if (id.equals(P_ID_SERIES)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_SERIES, P_SERIES));			
		} 
		
		// Volume
		if (id.equals(P_ID_VOLUME )) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_VOLUME, P_VOLUME));			
		}
		
		// Editor
		if (id.equals(P_ID_EDITOR)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_EDITOR, P_EDITOR));			
		} 

		// Edition
		if (id.equals(P_ID_EDITION)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_EDITION, P_EDITION));			
		} 
		
		// Place published
		if (id.equals(P_ID_PLACEPUBLISHED)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_PLACEPUBLISHED, P_PLACEPUBLISHED));			
		}
		
		// Publisher
		if (id.equals(P_ID_PUBLISHER)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_PUBLISHER, P_PUBLISHER));			
		} 
		
		// ISSN
		if (id.equals(P_ID_ISSN)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_ISSN, P_ISSN));			
		} 
		
		// Institution
		if (id.equals(P_ID_INSTITUTION)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_INSTITUTION, P_INSTITUTION));			
		} 
		
		// School
		if (id.equals(P_ID_SCHOOL)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_SCHOOL, P_SCHOOL));			
		} 
		
		// In series
		if (id.equals(P_ID_INSERIES)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_INSERIES, P_INSERIES));			
		} 
		
		// Series part
		if (id.equals(P_ID_SERIESPART)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_SERIESPART, P_SERIESPART));			
		} 
		
		// ISBN
		if (id.equals(P_ID_ISBN)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_ISBN, P_ISBN));
		}
		
		// Organization
		if (id.equals(P_ID_ORGANIZATION)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_ORGANIZATION, P_ORGANIZATION));			
		} 
		
		// Printed unit
		if (id.equals(P_ID_PRINTEDUNIT)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_PRINTEDUNIT, P_PRINTEDUNIT));			
		} 
		
		// In book
		if (id.equals(P_ID_INBOOK)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_INBOOK, P_INBOOK));			
		}
		
		// Bibtex entry type
		if (id.equals(P_ID_BIBTEX_ENTRYTYPE)) {
			descriptors.addElement(
					new ComboBoxPropertyDescriptor(P_ID_BIBTEX_ENTRYTYPE, P_BIBTEX_ENTRYTYPE, P_BIBTEX_ENTRYTYPE_MENU));			
		}
	}

	/**
	 * The default implementation of <code>ReferencePropertySource</code> will
	 * search all <code>ReferenceBase</code> classes.
	 * 
	 * @return
	 */
	protected int getSearchType() {
		return IReferenceSearch.BIBREF;
	}

	/**
	 * The default implementation of <code>ReferencePropertySource</code> will show
	 * a drop-down of reference types, but subclasses may want to show a non-editable
	 * text descriptor of the reference type. 
	 */
	protected PropertyDescriptor createReferenceTypeDescriptor() {
		return new ComboBoxPropertyDescriptor(
				P_ID_REFERENCETYPE, P_REFERENCETYPE, getReferenceTypeArray());
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#getEditableValue()
	 */
	public Object getEditableValue() {
		return CdmUtils.Nz(reference.getTitleCache());
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#getPropertyDescriptors()
	 */
	public IPropertyDescriptor[] getPropertyDescriptors() {
		return (IPropertyDescriptor[]) descriptors.toArray(
                new IPropertyDescriptor[descriptors.size()]);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#getPropertyValue(java.lang.Object)
	 */
	public Object getPropertyValue(Object id) {
				
		// Reference Type: get integer index of drop-down menu
		if (id.equals(P_ID_REFERENCETYPE)) {
			Class referenceClass = reference.getClass();
			int menuIndex = 0;
//			for (Class clazz : referenceTypeMap.keySet()) {
			for (Class clazz : getReferenceClassSet()) {
				if (clazz == referenceClass) {
					return menuIndex;
				}
				menuIndex++;
			}
			return 0;
		}

		// Editable cache
		if (id.equals(P_ID_EDITABLECACHE)) {
			return (CdmUtils.Nz(reference.getTitleCache()));
		}
		
		// Protect cache?
		if (id.equals(P_ID_PROTECT_CACHE)) {
			if (reference.isProtectedTitleCache()) {
				return CACHE_PROTECTED;
			} else {
				return CACHE_NOT_PROTECTED;
			}
		}
		
		// Author Team: for now, just edit the cache
		// TODO AuthorTeamPropertySource
		if (id.equals(P_ID_AUTHORTEAM)) {
			TeamOrPersonBase authorTeam = reference.getAuthorTeam();
			
			if (authorTeam == null) {
				return "";
			}
			
			return CdmUtils.Nz(authorTeam.getTitleCache());
		}
		
		// Year
		if (id.equals(P_ID_YEAR)) {
			return CdmUtils.Nz(reference.getYear());
		}
		
		// URI
		if (id.equals(P_ID_URI)) {
			return CdmUtils.Nz(reference.getUri());
		}
		
		// Date published
		if (id.equals(P_ID_DATEPUBLISHED)) {
			TimePeriod datePublished = (TimePeriod) invokeMethod(reference, "getDatePublished", null);
			
			if (datePublished == null) {
				datePublished = TimePeriod.NewInstance();
			}
			
			TimePeriodPropertySource timePeriodPropertySource = 
					new TimePeriodPropertySource(datePublished);
			timePeriodPropertySource.addPropertyChangeListener(new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent evt) {
					if (reference instanceof StrictReferenceBase && evt.getNewValue() instanceof TimePeriod) {	
						((StrictReferenceBase) reference).setDatePublished((TimePeriod) evt.getNewValue());
					}
				}
			});
			return timePeriodPropertySource;
		}
		
		// Title: used by StrictReferenceBase
		if (id.equals(P_ID_TITLE)) {
			String title = (String) invokeMethod(reference, "getTitle", null);
			return CdmUtils.Nz(title);
		}
		
		// InJournal: return an instance of ReferencePropertySource
		if (id.equals(P_ID_INJOURNAL)) {
				
			Journal inJournal = (Journal) invokeMethod(reference, "getInJournal", null);
			
			if (inJournal == null) {
				inJournal = Journal.NewInstance();
			}
			
			ReferencePropertySource bibRefPropertySource = new SingleRefTypePropertySource(
					inJournal, SingleRefTypePropertySource.JOURNAL);
			bibRefPropertySource.addPropertyChangeListener(new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent evt) {
					if (reference instanceof Article && evt.getNewValue() instanceof Journal) {	
						((Article) reference).setInJournal((Journal) evt.getNewValue());
					}
				}
			});
			return bibRefPropertySource;
		}
		
		// Pages
		if (id.equals(P_ID_PAGES)) {
			String pages = (String) invokeMethod(reference, "getPages", null);
			return CdmUtils.Nz(pages);
		}
		
		// Series
		if (id.equals(P_ID_SERIES)) {
			String series = (String) invokeMethod(reference, "getSeries", null);
			return CdmUtils.Nz(series); 
		}

		// Volume
		if (id.equals(P_ID_VOLUME)) {
			String volume = (String) invokeMethod(reference, "getVolume", null);
			return CdmUtils.Nz(volume);			
		}

		// Editor
		if (id.equals(P_ID_EDITOR)) {
			String editor = (String) invokeMethod(reference, "getEditor", null);
			return CdmUtils.Nz(editor);			
		}

		// Edition
		if (id.equals(P_ID_EDITION)) {
			String edition = (String) invokeMethod(reference, "getEdition", null);
			return CdmUtils.Nz(edition);			
		}
		
		// Place published
		if (id.equals(P_ID_PLACEPUBLISHED)) {
			String placePublished = (String) invokeMethod(reference, "getPlacePublished", null);
			return CdmUtils.Nz(placePublished);
		}
		
		// Publisher
		if (id.equals(P_ID_PUBLISHER)) {
			String publisher = (String) invokeMethod(reference, "getPublisher", null);
			return CdmUtils.Nz(publisher);			
		}

		// ISSN
		if (id.equals(P_ID_ISSN)) {
			String issn = (String) invokeMethod(reference, "getIssn", null);
			return CdmUtils.Nz(issn);			
		}

		// Institution
		if (id.equals(P_ID_INSTITUTION)) {
			String institution = (String) invokeMethod(reference, "getInstitution", null);
			return CdmUtils.Nz(institution);			
		}

		// In series
		if (id.equals(P_ID_INSERIES)) {
			
		}
		
		// ISBN
		if (id.equals(P_ID_ISBN)) {
			String isbn = (String) invokeMethod(reference, "getIsbn", null);
			return CdmUtils.Nz(isbn);				
		}

		// Organization
		if (id.equals(P_ID_ORGANIZATION)) {
			String organization = (String) invokeMethod(reference, "getOrganization", null);
			return CdmUtils.Nz(organization);			
		}
		
		// In book
		if (id.equals(P_ID_INBOOK)) {
			
			Book inBook = (Book) invokeMethod(reference, "getInBook", null);
			
			if (inBook == null) {
				inBook = Book.NewInstance();
			}
			
			ReferencePropertySource bibRefPropertySource = new SingleRefTypePropertySource(
					inBook, SingleRefTypePropertySource.BOOK);
			bibRefPropertySource.addPropertyChangeListener(new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent evt) {
					if (reference instanceof BookSection && evt.getNewValue() instanceof Book) {	
						((BookSection) reference).setInBook((Book) evt.getNewValue());
					}
				}
			});
			return bibRefPropertySource;
		}

		// Bibtex entry type
		if (id.equals(P_ID_BIBTEX_ENTRYTYPE)) {
			
		}
		
		return "";
	}

	/**
	 * Executes <code>methodName</code> on <code>obj</code> with
	 * the argument <code>arg</code>.
	 * <p>
	 * On failure, returns <code>null</code>.
	 * <p>
	 * NOTE:  both <code>Class.getMethod()</code> and <code>Method.invoke()</code>
	 * can be called with the <code>Object... args</code> approach, which allows
	 * the user to pass in a varying number of arguments. To keep it simple, and since
	 * this is usually only getters or setters, we are using only one argument.
	 * @param obj
	 * @param methodName
	 * @param args
	 * @return
	 */
	private Object invokeMethod(Object obj, String methodName, Object arg) {
		try {
			Method method = null;
			if (arg == null) {
				method = obj.getClass().getMethod(methodName);
				return method.invoke(obj);
			} else {
				method = obj.getClass().getMethod(methodName, arg.getClass());
				return method.invoke(obj, arg);
			}
		} catch (SecurityException e) {
			logger.error(e.getMessage());
		} catch (IllegalArgumentException e) {
			logger.error(e.getMessage());
		} catch (NoSuchMethodException e) {
			logger.error(e.getMessage());
		} catch (IllegalAccessException e) {
			logger.error(e.getMessage());
		} catch (InvocationTargetException e) {
			logger.error(e.getMessage());
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#isPropertySet(java.lang.Object)
	 */
	public boolean isPropertySet(Object id) {
		return false;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#resetPropertyValue(java.lang.Object)
	 */
	public void resetPropertyValue(Object id) {}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#setPropertyValue(java.lang.Object, java.lang.Object)
	 */
	public void setPropertyValue(Object id, Object value) {
		
		// Search result
		if (id.equals(P_ID_SEARCH)) {
			if (value instanceof ReferenceBase) {
				reference = (ReferenceBase) value;
			}
		}

		// Protect cache? 
		if (id.equals(P_ID_PROTECT_CACHE)) {
			if (((Integer) value).intValue() == CACHE_PROTECTED) {
				reference.setProtectedTitleCache(true);
			} else {
				reference.setProtectedTitleCache(false);
			}
		}
				
		// Reference Type
		if (id.equals(P_ID_REFERENCETYPE)) {
			
			// NOTE: this code will only be reached if user actually changes choice in reference type
			//	dropdown list, not if focus is merely gained and lost without changing the value
			
			// Grab reference's title cache
			String oldTitleCache = reference.getTitleCache();
			
			// Get reference's class
			int menuIndex;
			try {
				menuIndex = ((Integer) value).intValue();
			} catch (ClassCastException e) {
				// Leave method / no change on class cast error
				logger.error(e.getMessage());
				return;
			}			
			Class referenceClass = (Class)(getReferenceClassSet().toArray())[menuIndex];
			
			// Make a new reference using the class
			reference = makeNewReference(referenceClass);
			
			// Set new reference's title cache
			reference.setTitleCache(oldTitleCache);
		}
		
		// Author Team
		if (id.equals(P_ID_AUTHORTEAM)) {
			TeamOrPersonBase authorTeam = reference.getAuthorTeam();
			if (authorTeam == null) {
				authorTeam = Team.NewInstance();
				reference.setAuthorTeam(authorTeam);
			}
			try {
				authorTeam.setTitleCache((String) value);
			} catch (ClassCastException e) {
				// Leave method / no change on class cast error
				logger.error(e.getMessage());
				return;
			}
		}
		
		// Citation
		if (id.equals(P_ID_URI)) {
			String uri = (String) value;
			reference.setUri(uri);
		}
		
		// Year - only set directly by Bibtex refs
		if (id.equals(P_ID_YEAR)) {
			String year = (String) value;
			if (reference instanceof BibtexReference) {
				((BibtexReference) reference).setYear(year);
			}
		}

		// Title: used by StrictReferenceBase
		if (id.equals(P_ID_TITLE)) {
			String title = null;
			try {
				title = (String) value;
			} catch (ClassCastException e) {
				// Leave method / no change on class cast error
				logger.error(e.getMessage());
				return;
			}
			invokeMethod(reference, "setTitle", title);
		}

		// Pages
		if (id.equals(P_ID_PAGES)) {
			invokeMethod(reference, "setPages", (String) value);
		}

		// Series
		if (id.equals(P_ID_SERIES)) {
			invokeMethod(reference, "setSeries", (String) value);
		}
		
		// Volume
		if (id.equals(P_ID_VOLUME)) {
			invokeMethod(reference, "setVolume", (String) value);
		}
		
		// Editor
		if (id.equals(P_ID_EDITOR)) {
			invokeMethod(reference, "setEditor", (String) value);
		}

		// Edition
		if (id.equals(P_ID_EDITION)) {
			invokeMethod(reference, "setEdition", (String) value);
		}
		
		// Place published
		if (id.equals(P_ID_PLACEPUBLISHED)) {
			invokeMethod(reference, "setPlacePublished", (String) value);
		}

		// Publisher
		if (id.equals(P_ID_PUBLISHER)) {
			invokeMethod(reference, "setPublisher", (String) value);
		}

		// ISSN
		if (id.equals(P_ID_ISSN)) {
			invokeMethod(reference, "setIssn", (String) value);
		}

		// Institution
		if (id.equals(P_ID_INSTITUTION)) {
			invokeMethod(reference, "setInstitution", (String) value);			
		}
		
		// ISBN
		if (id.equals(P_ID_ISBN)) {
			invokeMethod(reference, "setIsbn", (String) value);
		}

		// Organization
		if (id.equals(P_ID_ORGANIZATION)) {
			invokeMethod(reference, "setOrganization", (String) value);
		}

		// In book
		if (id.equals(P_ID_INBOOK)) {
			Book inBook = (Book) invokeMethod(reference, "getInBook", null);
			if (inBook == null) {
				inBook = Book.NewInstance();
				invokeMethod(reference, "setInBook", inBook);
			}
			inBook.setTitleCache((String) value);
		}
		
		// Date published
		if (id.equals(P_ID_DATEPUBLISHED)) {
			
			if (!(reference instanceof StrictReferenceBase)) {
				return;
			}
			
			if (value instanceof String) {
				String year = (String) value;
				TimePeriod yearsPublished = null;
				
				// If not empty, format is either "xxxx" or "xxxx-xxxx", thanks to validator
				if (!year.equals("")) {
					
					// Convert to a TimePeriod
					yearsPublished = CdmUtil.convertTimePeriod(year);
					
					// Update reference's dataPublished field
					TimePeriod datePublished = ((StrictReferenceBase) reference).getDatePublished();
					
					if (datePublished == null) {
						((StrictReferenceBase) reference).setDatePublished(yearsPublished);
					} else {
						
						// Any months or days should be left as is - only update years 
						datePublished.setStartYear(yearsPublished.getStartYear());
						datePublished.setEndYear(yearsPublished.getEndYear());
					}
				} else {
					((StrictReferenceBase) reference).setDatePublished(null);
				}
				

			}
		}

		// Editable cache
		if (id.equals(P_ID_EDITABLECACHE)) {
			reference.setTitleCache((String) value);
		} else {
			reference.setTitleCache(reference.generateTitle(), reference.isProtectedTitleCache());
		}
		
		propertyChangeSupport.firePropertyChange(ITaxEditorConstants.PROPERTY_SHEET_CHANGE, null, reference);
	}
	
	/**
	 * Takes a <code>referenceClass</code>, then executes either its <code>NewInstance()</code> 
	 * method or its constructor. Puts the resulting object in <code>this.reference</code> and
	 * returns it.
	 * <p>
	 * If unsuccessful, <code>this.reference</code> keeps its old value and is returned.
	 * @param referenceClass
	 * @return
	 */
	private ReferenceBase makeNewReference(Class referenceClass) {
				
		// Check whether the class has a method "NewInstance()";
		try {
			Method method = referenceClass.getMethod("NewInstance", null);
			reference = (ReferenceBase) method.invoke(null, null);
		} catch (NoSuchMethodException e) {
			try {
				// If not, use its constructor
				Constructor constructor = referenceClass.getConstructor(null);
				reference = (ReferenceBase) constructor.newInstance(null);
			} catch (SecurityException e1) {
				e1.printStackTrace();
			} catch (NoSuchMethodException e1) {
				e1.printStackTrace();
			} catch (IllegalArgumentException e1) {
				e1.printStackTrace();
			} catch (InstantiationException e1) {
				e1.printStackTrace();
			} catch (IllegalAccessException e1) {
				e1.printStackTrace();
			} catch (InvocationTargetException e1) {
				e1.printStackTrace();
			}
		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		}
		
		return reference;
	}

	public String toString() {
		return CdmUtils.Nz(reference.getTitleCache());
	}

	private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);
	
	public void addPropertyChangeListener(
			PropertyChangeListener listener) {
		propertyChangeSupport.addPropertyChangeListener(listener);
	}
}